{$INCLUDE ..\cDefines.inc}
{$DEBUGINFO ON}
unit cTCPSocks5ListClient;

interface

uses
  { Delphi }
  Windows,
  WinSock,
  SysUtils,
  Classes,

  { Fundamentals }
  cTCPClient;

type
  TfndTCPSocks5ListClient = class(ATCPClient)
  private
    FChanging,FChanged:boolean;
    FSocksList:TStringList;
    function GetSocksList: TStrings;
    procedure SetSocksList(const Value: TStrings);
    procedure SockListChanging(Sender: TObject);
    procedure SockListChange(Sender: TObject);
    procedure CreateSocks(Sender:TStringList);
    procedure MakeSocksChain(Sender:TStringList);
    function GetProxiesChainCreated: boolean;
    procedure SetProxiesChainCreated(const Value: boolean);
  public
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;
  published
    property  Host;
    property  Port;

    property  LocalHost;
    property  LocalPort;

    property  ProxiesChainCreated:boolean read GetProxiesChainCreated write SetProxiesChainCreated;
    property  Socks5List:TStrings read GetSocksList write SetSocksList;

    property  StreamMode;
    property  TimeOut;
    property  RunInThread;

    property  Collection;
    property  OnLog;
    property  LogTo;

    property  Active;

    property  ThrottleWrite;
    property  WriteThrottleRate;
    property  ThrottleRead;
    property  ReadThrottleRate;

    property  OnStateChange;
    property  OnSocketStateChange;
    property  OnNegotiating;
    property  OnConnected;
    property  OnConnectFailed;
    property  OnDataAvailable;
    property  OnClose;
    property  OnThreadRun;
    property  OnThreadRunComplete;

    property  OnSyncSocketStateChange;
    property  OnSyncNegotiating;
    property  OnSyncConnected;
    property  OnSyncConnectFailed;
    property  OnSyncDataAvailable;
    property  OnSyncClose;
  end;

procedure Register;

implementation

uses
  // Delphi
  Messages,
  // Fundamentals
  cUtils;

procedure Register;
begin
  RegisterComponents('fnd Sockets', [TfndTCPSocks5ListClient]);
end;

{ TfndTCPSocks5ListClient }

constructor TfndTCPSocks5ListClient.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FChanging:=false;
  FChanged:=false;
  FSocksList:=TStringList.Create;
  FSocksList.OnChange:=SockListChange;
  FSocksList.OnChanging:=SockListChanging;
end;

procedure TfndTCPSocks5ListClient.CreateSocks(Sender: TStringList);
var i:integer;
begin
  for i:=0 to Sender.Count-1 do
  if Sender.Objects[i]=nil then
    Sender.Objects[i]:=TfndSocks5Proxy.Create(Self);
end;

destructor TfndTCPSocks5ListClient.Destroy;
begin
  FSocksList.Free;
  inherited;
end;

function TfndTCPSocks5ListClient.GetProxiesChainCreated: boolean;
begin
  result:=assigned(Proxy);
end;

function TfndTCPSocks5ListClient.GetSocksList: TStrings;
begin
  result:=FSocksList;
end;

procedure TfndTCPSocks5ListClient.MakeSocksChain(Sender: TStringList);
var i:integer;
    s:ATCPClientConnectionProxy;

  function GetHost(s:string):string;
  var i:integer;
  begin
    i:=pos(':',s);
    if i=1 then
    begin
      result:='127.0.0.1';
      exit;
    end;
    if i>0 then
      result:=copy(s,1,i-1)
    else
      result:=s;
  end;

  function GetPort(s:string):string;
  var i:integer;
  begin
    i:=pos(':',s);
    if i=0 then
    begin
      result:='1080';
      exit;
    end;
    if i>0 then
    if i<length(s) then
      result:=copy(s,i+1,maxint)
    else
      result:='1080';
  end;

begin
  for i:=0 to Sender.Count-1 do
  begin
    if Sender.Objects[i]=nil then raise Exception.Create('Socks chain tears!');
    s:=ATCPClientConnectionProxy(Sender.Objects[i]);
    s.Host:=GetHost(Sender[i]);
    s.Port:=GetPort(Sender[i]);
    if i>0 then
      ATCPClientConnectionProxy(Sender.Objects[i-1]).NextProxy:=s;
  end;
  if Sender.Count>0 then
    Self.Proxy:=ATCPClientConnectionProxy(Sender.Objects[0])
  else
    Self.Proxy:=nil;
end;

procedure TfndTCPSocks5ListClient.SetProxiesChainCreated(
  const Value: boolean);
begin
//nothing to do
end;

procedure TfndTCPSocks5ListClient.SetSocksList(const Value: TStrings);
begin
  FSocksList.Assign(value);
end;

procedure TfndTCPSocks5ListClient.SockListChange(Sender: TObject);
begin
  if FChanged or FChanging then exit;
  FChanged:=true;
  CreateSocks(TStringList(Sender));
  MakeSocksChain(TStringList(Sender));
  FChanged:=false;
end;

procedure TfndTCPSocks5ListClient.SockListChanging(Sender: TObject);
var i:integer;
    str:TStringList;
begin
  if FChanged or FChanging then exit;
  str:=TStringList(Sender);
  FChanging:=true;
  for i:=0 to str.Count-1 do
  if str.Objects[i]<>nil then
  begin
    TfndSocks5Proxy(str.Objects[i]).Free;
    str.Objects[i]:=nil;
  end;
  FChanging:=false;
end;

end.

